(*****************************************************************************
 *  Author:
 *    Daniel N. Gtzmann <dngoetzmann@googlemail.com>
 *
 *  Copyright:
 *     Daniel N. Gtzmann, 2009
 *
 *  Last modified:
 *    $Date: 2009-01-30 16:26:36 +0100 (Fri, 30 Jan 2009) $
 *    $Author: goetzmann $
 *    $Revision: 373 $
 *
 *  This file is part of Spartacus,
 *  the tableau prover for hybrid logic
 *
 *  Permission is hereby granted, free of charge, to any person obtaining
 *  a copy of this software and associated documentation files (the
 *  "Software"), to deal in the Software without restriction, including
 *  without limitation the rights to use, copy, modify, merge, publish,
 *  distribute, sublicense, and/or sell copies of the Software, and to
 *  permit persons to whom the Software is furnished to do so, subject to
 *  the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 *  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 *  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 *  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *****************************************************************************)


structure Parsetree =
	struct
		datatype parsetree =
			  CONJ of parsetree list
			| DISJ of parsetree list
			| NEG of parsetree
			| PROPVAR of string
			| NOMINAL of string
			| DIAMOND of string * parsetree
			| BOX of string * parsetree
			| ALL of parsetree
			| EXISTS of parsetree
			| DIFF of parsetree
			| NEGDIFF of parsetree
			| AT of string * parsetree
		
		
		fun neg (NEG x) = x
		  | neg x = NEG x
		
		
		fun conj (CONJ xs) (CONJ ys) = CONJ (xs @ ys)
		  | conj (CONJ xs) y         = CONJ (xs @ [y])
		  | conj x         (CONJ ys) = CONJ (x::ys)
		  | conj x         y         = CONJ [x, y]
		
		
		fun disj (DISJ xs) (DISJ ys) = DISJ (xs @ ys)
		  | disj (DISJ xs) y         = DISJ (xs @ [y])
		  | disj x         (DISJ ys) = DISJ (x::ys)
		  | disj x         y         = DISJ [x, y]
		
		
		fun impl x y = DISJ [neg x, y]
		
		
		fun xor x y = DISJ [CONJ [neg x, y], CONJ [x, neg y]]
		
		
		fun mxor nil = DISJ nil
		  | mxor [x] = x
		  | mxor (x::xr) = xor x (mxor xr)
		
		
		fun mequal xs = DISJ [CONJ xs, CONJ (map NEG xs)]
		
		
		fun toString (CONJ ts) =
				"/\\(" ^ (Util.listToString toString ts) ^ ")"
		  | toString (DISJ ts) =
				"\\/(" ^ (Util.listToString toString ts) ^ ")"
		  | toString (NEG t) = "~" ^ (toString t)
		  | toString (PROPVAR p) = p
		  | toString (NOMINAL n) = "(=)" ^ n
		  | toString (DIAMOND (r, t)) = "<" ^ r ^ ">" ^ (toString t)
		  | toString (BOX (r, t)) = "[" ^ r ^ "]" ^ (toString t)
		  | toString (ALL t) = "A" ^ (toString t)
		  | toString (EXISTS t) = "E" ^ (toString t)
		  | toString (DIFF t) = "D" ^ (toString t)
		  | toString (NEGDIFF t) = "(~D)" ^ (toString t)
		  | toString (AT (x, t)) = "@" ^ x ^ ":" ^ (toString t)
		
		
		fun flatten (PROPVAR p) = (PROPVAR p)
		  | flatten (NOMINAL n) = (NOMINAL n)
		  | flatten (NEG (NEG t)) = t
		  | flatten (NEG t) = NEG (flatten t)
		  | flatten (CONJ ts) =
				let
					val (cnj, oth) = List.partition (fn CONJ _ => true | _ => false) (map flatten ts)
					
					fun ff (CONJ ts, xs) = ts @ xs
					  | ff _ = Exn.unexpected "Parsetree.flatten.ff: #1 of arg is not a CONJ _"
				in
					CONJ (oth @ (foldl ff nil cnj))
				end
		  | flatten (DISJ ts) =
				let
					val (dsj, oth) = List.partition (fn DISJ _ => true | _ => false) (map flatten ts)
					
					fun ff (DISJ ts, xs) = ts @ xs
					  | ff _ = Exn.unexpected "Parsetree.flatten.ff: #1 of arg is not a DISJ _"
				in
					DISJ (oth @ (foldl ff nil dsj))
				end
		  | flatten (DIAMOND (r, t)) = DIAMOND (r, flatten t)
		  | flatten (BOX (r, t)) = BOX (r, flatten t)
		  | flatten (ALL t) = ALL (flatten t)
		  | flatten (EXISTS t) = EXISTS (flatten t)
		  | flatten (DIFF t) = DIFF (flatten t)
		  | flatten (NEGDIFF t) = NEGDIFF (flatten t)
		  | flatten (AT (x, t)) = AT (x, flatten t)
		
		
		fun nnf t =
			let
				fun nnf' (PROPVAR p) = PROPVAR p
				  | nnf' (NOMINAL n) = NOMINAL n
				  | nnf' (NEG (PROPVAR p)) = NEG (PROPVAR p)
				  | nnf' (NEG (NOMINAL n)) = NEG (NOMINAL n)
				  | nnf' (CONJ ts) = CONJ (map nnf' ts)
				  | nnf' (NEG (CONJ ts)) = DISJ (map (nnf' o NEG) ts)
				  | nnf' (DISJ ts) = DISJ (map nnf' ts)
				  | nnf' (NEG (DISJ ts)) = CONJ (map (nnf' o NEG) ts)
				  | nnf' (DIAMOND (r, t)) = DIAMOND (r, nnf' t)
				  | nnf' (NEG (DIAMOND (r, t))) = BOX (r, nnf' (NEG t))
				  | nnf' (BOX (r, t)) = BOX (r, nnf' t)
				  | nnf' (NEG (BOX (r, t))) = DIAMOND (r, nnf' (NEG t))
				  | nnf' (ALL t) = ALL (nnf' t)
				  | nnf' (NEG (ALL t)) = EXISTS (nnf' (NEG t))
				  | nnf' (EXISTS t) = EXISTS (nnf' t)
				  | nnf' (NEG (EXISTS t)) = ALL (nnf' (NEG t))
				  | nnf' (DIFF t) = DIFF (nnf' t)
				  | nnf' (NEG (DIFF t)) = NEGDIFF (nnf' (NEG t))
				  | nnf' (NEGDIFF t) = NEGDIFF (nnf' t)
				  | nnf' (NEG (NEGDIFF t)) = DIFF (nnf' (NEG t))
				  | nnf' (AT (x, t)) = AT (x, nnf' t)
				  | nnf' (NEG (AT (x, t))) = AT (x, nnf' (NEG t))
				  | nnf' (NEG (NEG t)) = nnf' t
			in
				flatten (nnf' t)
			end
		
		fun toNumberedTree pt =
			let
				exception NotFound
				
				datatype varType = NV | PV | RV
				
				val bm = ref (Binarymap.mkDict String.compare) : (string, int) Binarymap.dict ref
				
				val propvars = ref IntBinarySet.empty
				val nominals = ref IntBinarySet.empty
				val relvars = ref IntBinarySet.empty
				
				fun lookup vt v =
					let
						val prefix =
							case vt
								of NV => "nv"
								|  PV => "pv"
								|  RV => "rv"
						
						fun assignNew () =
							let
								fun keepDigits x = if Char.isDigit x then Char.toString x else ""
								
								val number = 
									let
										fun setNext s n =
											if IntBinarySet.member (!s, n)
											then setNext s (n + 1)
											else (s := (IntBinarySet.add (!s, n)); n)
										
										val s =
											case vt
												of NV => nominals
												|  PV => propvars
												|  RV => relvars
										
										val n = Option.getOpt (Int.fromString (String.translate keepDigits v), 1)
									in
										setNext s (Int.max (n, 1))
									end
							in
								(Ref.modify (fn m => Binarymap.insert (m, prefix ^ v, number)) bm; number)
							end
					in
						case Binarymap.peek (!bm, prefix ^ v)
							of SOME v' => Int.toString v'
							|  NONE => Int.toString (assignNew ())
					end
				
				fun tnt (CONJ ts) = CONJ (map tnt ts)
				  | tnt (DISJ ts) = DISJ (map tnt ts)
				  | tnt (NEG t) = NEG (tnt t)
				  | tnt (PROPVAR p) = PROPVAR (lookup PV p)
				  | tnt (NOMINAL p) = NOMINAL (lookup NV p)
				  | tnt (DIAMOND (r, t)) = DIAMOND (lookup RV r, tnt t)
				  | tnt (BOX (r, t)) = BOX (lookup RV r, tnt t)
				  | tnt (ALL t) = ALL (tnt t)
				  | tnt (EXISTS t) = EXISTS (tnt t)
				  | tnt (DIFF t) = DIFF (tnt t)
				  | tnt (NEGDIFF t) = NEGDIFF (tnt t)
				  | tnt (AT (n, t)) = AT (lookup NV n, tnt t)
			in
				tnt pt
			end
		
		
		fun toCnf pt =
			let
				fun toCnf' (PROPVAR p) = PROPVAR p
				  | toCnf' (NEG (PROPVAR p)) = NEG (PROPVAR p)
				  | toCnf' (CONJ ts) = flatten (CONJ (map toCnf' ts))
				  | toCnf' (DISJ ts) =
					let
						fun distribute (DISJ ts) =
							let
								val (conj, lit) = List.partition (fn CONJ _ => true | _ => false) ts
								
								fun prod (xs, nil) = nil
								  | prod (xs, (y::yr)) = (map (fn x => y::x) xs) @ (prod (xs, yr))
								
								fun distribute' ds nil = ds
								  | distribute' ds ((CONJ ts')::xr) = distribute' (prod (ds, ts')) xr
								  | distribute' _ _ = Exn.unexpArg "Parsetree.toCnf.toCnf'.distribute.distribute'"
							in
								if conj = nil
								then DISJ lit
								else CONJ (map DISJ (distribute' [lit] conj))
							end
						  | distribute _ = Exn.unexpArg "Parsetree.toCnf.toCnf'.distribute"
					in
						distribute (flatten (DISJ (map toCnf' ts)))
					end
				  | toCnf' _ = Exn.error "Unable to transform formula to CNF"
			in
				case toCnf' (nnf pt)
					of CONJ ts => CONJ (map (fn t => case t of x as DISJ _ => x | t' => DISJ [t']) ts)
					 | DISJ ts => CONJ [DISJ ts]
					 | t => CONJ  [DISJ [t]]
			end
	end
